// n-body problem, OpenCL solution
//
// compile:
//      module load CUDA
//      gcc nbody-cl.c -fopenmp -lOpenCL -o nbody-cl
//
// usage:
//      module load CUDA
//      ./nbody-cl <bodies> <time-steps> <kernel id> <workgroup-size>
//
//      module load CUDA
//      srun --reservation=fri --constrain=intel --gpus=1 ./nbody-cl <bodies> <time-steps> <kernel id> <workgroup-size>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>
#include "CL/cl.h"
#include "nbody.h"

#define MAX_SOURCE_SIZE 16384

#define REAL3 MACROCONCAT(cl_, REAL, 3)
#define REAL4 MACROCONCAT(cl_, REAL, 4)

int nBody_init(REAL4 *b, REAL3 *v, int bodies)
{
    srand(123456);
    for(int i = 0; i < bodies; i++)
    {
        b[i].w = 1.0;
        b[i].x = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        b[i].y = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        b[i].z = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        v[i].x = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        v[i].y = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        v[i].z = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
    }

    return 0;
}

int nBody_state(REAL4 *b, int bodies)
{
    REAL3 pos;
    
    pos.x = 0.0;
    pos.y = 0.0;
    pos.z = 0.0;

    for (int i = 0; i < bodies; i++)
    {
        pos.x += b[i].x;
        pos.y += b[i].y;
        pos.z += b[i].z;
    }
    pos.x /= bodies;
    pos.y /= bodies;
    pos.z /= bodies;

    printf("Position of %d bodies: x = %lf, y = %lf, z = %lf\n", bodies, pos.x, pos.y, pos.z);

    return 0;
}

int main(int argc, char * argv[])
{
    if (argc < 5)
    {
        printf("%s <bodies> <time-steps> <kernel id> <workgroup-size>\n", argv[0]);
        return 1;
    }

    // read arguments
    int bodies = atoi(argv[1]);
    int timesteps = atoi(argv[2]);
    char fileName[100];
    sprintf(fileName, "%s-%s.cl", argv[0], argv[3]);
    int workgroup_size = atoi(argv[4]);

    // allocate data structures
    REAL4 *b = (REAL4 *)malloc(bodies*sizeof(REAL4));
    REAL3 *v = (REAL3 *)malloc(bodies*sizeof(REAL3));
    REAL3 *a = (REAL3 *)malloc(bodies*sizeof(REAL3));

    // initialization
    nBody_init(b, v, bodies);
    nBody_state(b, bodies);

    // OpenCL
    // Get platforms
    cl_uint num_platforms;
    cl_int clStatus = clGetPlatformIDs(0, NULL, &num_platforms);
    if(num_platforms == 0)
    {
        printf("No OpenCL platforms\n");
        return 1;
    }
    cl_platform_id *platforms = (cl_platform_id *)malloc(sizeof(cl_platform_id)*num_platforms);
    clStatus = clGetPlatformIDs(num_platforms, platforms, NULL);
    // Get devices for selected platform
    cl_uint num_devices;
    clStatus = clGetDeviceIDs(platforms[0], CL_DEVICE_TYPE_GPU, 0, NULL, &num_devices);
    if(num_devices == 0)
    {
        printf("No OpenCL devices for selected platform.\n");
        return 1;
    }
    num_devices = 1; // limit to one device
    cl_device_id *devices = (cl_device_id *)malloc(sizeof(cl_device_id)*num_devices);
    clStatus = clGetDeviceIDs(platforms[0], CL_DEVICE_TYPE_GPU, num_devices, devices, NULL);
    // Create context for one device
    cl_context context = clCreateContext(NULL, num_devices, devices, NULL, NULL, &clStatus);
    // Create a command queue (0: in order, 1: out of order)
    cl_command_queue command_queue = clCreateCommandQueue(context, devices[0], 0, &clStatus);
    // Read kernel from file
    FILE *fp;
    char *source_str;
    size_t source_size;
    fp = fopen(fileName, "r");
    if (!fp) 
	{
		fprintf(stderr, ":-(#\n");
        exit(1);
    }
    source_str = (char*)malloc(MAX_SOURCE_SIZE);
    source_size = fread(source_str, 1, MAX_SOURCE_SIZE, fp);
	source_str[source_size] = '\0';
    fclose( fp );
    // Create and build a program
    cl_program program = clCreateProgramWithSource(context,	1, (const char **)&source_str, NULL, &clStatus);
    clStatus = clBuildProgram(program, 1, devices, NULL, NULL, NULL);
	size_t build_log_len;
	char *build_log;
	clStatus = clGetProgramBuildInfo(program, devices[0], CL_PROGRAM_BUILD_LOG, 0, NULL, &build_log_len);
    if (build_log_len > 2)
    {
        build_log =(char *)malloc(sizeof(char)*(build_log_len+1));
        clStatus = clGetProgramBuildInfo(program, devices[0], CL_PROGRAM_BUILD_LOG, 
                                        build_log_len, build_log, NULL);
        printf("%s", build_log);
        free(build_log);
        return 1;
    }
    // Create memory buffers on the device
    cl_mem b_d = clCreateBuffer(context, CL_MEM_READ_WRITE, bodies * sizeof(REAL4), NULL, &clStatus);
    cl_mem v_d = clCreateBuffer(context, CL_MEM_READ_WRITE, bodies * sizeof(REAL3), NULL, &clStatus);
    cl_mem a_d = clCreateBuffer(context, CL_MEM_READ_WRITE, bodies * sizeof(REAL3), NULL, &clStatus);
	// Divide work
    size_t local_item_size = workgroup_size;
	size_t num_groups = ((bodies-1)/local_item_size+1);		
    size_t global_item_size = num_groups*local_item_size;		
    // create kernels and set arguments
    REAL eps2 = EPS2;
    REAL kappa = KAPPA;
    cl_kernel kernelInteractionBodyAll = clCreateKernel(program, "interactionBodyAll", &clStatus);
    clStatus  = clSetKernelArg(kernelInteractionBodyAll, 0, sizeof(cl_mem), (void *)&a_d);
    clStatus |= clSetKernelArg(kernelInteractionBodyAll, 1, sizeof(cl_mem), (void *)&b_d);
    clStatus |= clSetKernelArg(kernelInteractionBodyAll, 2, sizeof(cl_int), (void *)&bodies);
	clStatus |= clSetKernelArg(kernelInteractionBodyAll, 3, sizeof(REAL), (void *)&eps2);
	clStatus |= clSetKernelArg(kernelInteractionBodyAll, 4, sizeof(REAL), (void *)&kappa);
	clStatus |= clSetKernelArg(kernelInteractionBodyAll, 5, local_item_size*sizeof(REAL4), NULL);

    REAL timestep = TIMESTEP;
    cl_kernel kernelUpdateBody = clCreateKernel(program, "updateBody", &clStatus);
    clStatus  = clSetKernelArg(kernelUpdateBody, 0, sizeof(cl_mem), (void *)&b_d);
    clStatus |= clSetKernelArg(kernelUpdateBody, 1, sizeof(cl_mem), (void *)&v_d);
    clStatus |= clSetKernelArg(kernelUpdateBody, 2, sizeof(cl_mem), (void *)&a_d);
	clStatus |= clSetKernelArg(kernelUpdateBody, 3, sizeof(cl_int), (void *)&bodies);
	clStatus |= clSetKernelArg(kernelUpdateBody, 4, sizeof(REAL), (void *)&timestep);

    // simulation: start
    double dt = omp_get_wtime();    
    // Write data to device
    clStatus = clEnqueueWriteBuffer(command_queue, b_d, CL_TRUE, 0, bodies * sizeof(REAL4), b, 0, NULL, NULL);
    clStatus = clEnqueueWriteBuffer(command_queue, v_d, CL_TRUE, 0, bodies * sizeof(REAL3), v, 0, NULL, NULL);
    double dti = omp_get_wtime();    
    // do iterations
    for (int t = 0; t < timesteps; t++)
    {
        clStatus = clEnqueueNDRangeKernel(command_queue, kernelInteractionBodyAll, 1, NULL,						
                                          &global_item_size, &local_item_size, 0, NULL, NULL);	
        clStatus = clEnqueueNDRangeKernel(command_queue, kernelUpdateBody, 1, NULL,						
                                          &global_item_size, &local_item_size, 0, NULL, NULL);	
    }    
    // Wait for completion
    clStatus = clFlush(command_queue);
    clStatus = clFinish(command_queue);
    dti = omp_get_wtime() - dti;
    // Read data from device
    clStatus = clEnqueueReadBuffer(command_queue, b_d, CL_TRUE, 0, bodies * sizeof(REAL4), b, 0, NULL, NULL);
    clStatus = clEnqueueReadBuffer(command_queue, v_d, CL_TRUE, 0, bodies * sizeof(REAL3), v, 0, NULL, NULL);
    dt = omp_get_wtime() - dt;
    // simulation: end

    // output
    nBody_state(b, bodies);
    printf("Kernel = %s, WGsize = %d, ", argv[3], workgroup_size);
    printf("Time = %lf, Time per time-step = %lf\n\n", dt, dti / timesteps);

    // Release OpenCL objects
    clStatus = clReleaseKernel(kernelInteractionBodyAll);
    clStatus = clReleaseKernel(kernelUpdateBody);
    clStatus = clReleaseProgram(program);
    clStatus = clReleaseMemObject(b_d);
    clStatus = clReleaseMemObject(v_d);
    clStatus = clReleaseMemObject(a_d);
    clStatus = clReleaseCommandQueue(command_queue);
    clStatus = clReleaseContext(context);

    // Release Host objects
    free(b);
    free(v);
    free(a);

    return 0;    
}
