// n-body problem, OpenMP solution
//
// compile: 
//      gcc nbody-mp.c -lm -fopenmp -o nbody-mp
//
// usage:
//      export OMP_NUM_THREADS=<threads>
//      ./nbody-mp <bodies> <time-steps>
//
//      export OMP_NUM_THREADS=<threads>
//      srun --reservation=fri --constraint=intel --cpus-per-task=<threads> ./nbody-mp 2000 100


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "omp.h"
#include "nbody.h"

struct REAL3    // coordinates
{
    REAL x;
    REAL y;
    REAL z;
};

struct REAL4    // body (coordinates + mass) 
{
    REAL x;     // coordinates
    REAL y;
    REAL z;
    REAL w;     // mass
};

int interactionBodyBody(struct REAL3 *ai, struct REAL4 bi, struct REAL4 bj, REAL eps2, REAL kappa) 
{
    struct REAL3 r;
    
    r.x = bi.x - bj.x;
    r.y = bi.y - bj.y;
    r.z = bi.z - bj.z;

    REAL r2 = r.x * r.x + r.y * r.y + r.z * r.z + eps2; // r^2
    REAL rinv = 1/sqrt(r2);                             // r^{-1}
	REAL a_rinv = kappa * bj.w * rinv * rinv * rinv;    // a r^{-1}

    (*ai).x += a_rinv * r.x;
    (*ai).y += a_rinv * r.y;
    (*ai).z += a_rinv * r.z;

    return 0;
}

int interactionBodyAll(struct REAL3 *ai, struct REAL4 bi, struct REAL4 *b, int bodies, REAL eps2, REAL kappa)
{
    (*ai).x = (*ai).y = (*ai).z = 0.0;

    for (int j = 0; j < bodies; j++)
        interactionBodyBody(ai, bi, b[j], eps2, kappa);

    return 0;
}

int updateBody(struct REAL4 *bi, struct REAL3 *vi, struct REAL3 ai, REAL dt)
{
    (*bi).x += (*vi).x * dt + 0.5 * ai.x * dt * dt;
    (*bi).y += (*vi).y * dt + 0.5 * ai.y * dt * dt;
    (*bi).z += (*vi).z * dt + 0.5 * ai.z * dt * dt;

    (*vi).x += ai.x * dt;
    (*vi).y += ai.y * dt;
    (*vi).z += ai.z * dt;

    return 0;
}

int nBody_init(struct REAL4 *b, struct REAL3 *v, int bodies)
{
    srand(123456);
    for(int i = 0; i < bodies; i++)
    {
        b[i].w = 1.0;
        b[i].x = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        b[i].y = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        b[i].z = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        v[i].x = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        v[i].y = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        v[i].z = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
    }

    return 0;
}

int nBody_state(struct REAL4 *b, int bodies)
{
    struct REAL3 pos;
    
    pos.x = 0.0;
    pos.y = 0.0;
    pos.z = 0.0;

    for (int i = 0; i < bodies; i++)
    {
        pos.x += b[i].x;
        pos.y += b[i].y;
        pos.z += b[i].z;
    }
    pos.x /= bodies;
    pos.y /= bodies;
    pos.z /= bodies;

    printf("Position of %d bodies: x = %lf, y = %lf, z = %lf\n", bodies, pos.x, pos.y, pos.z);

    return 0;
}

int main(int argc, char * argv[])
{
    if (argc < 3)
    {
        printf("%s <bodies> <time-steps>\n", argv[0]);
        return 1;
    }

    int bodies = atoi(argv[1]);
    int timesteps = atoi(argv[2]);

    // allocate structures,
    struct REAL4 *b = (struct REAL4 *)malloc(bodies*sizeof(struct REAL4));
    struct REAL3 *v = (struct REAL3 *)malloc(bodies*sizeof(struct REAL3));
    struct REAL3 *a = (struct REAL3 *)malloc(bodies*sizeof(struct REAL3));

    // initialization
    nBody_init(b, v, bodies);
    nBody_state(b, bodies);

    // simulation
    #pragma omp barrier
    double dt = omp_get_wtime();    
    for (int t = 0; t < timesteps; t++)
    {
        #pragma omp parallel for
        for(int i = 0; i < bodies; i++)
            interactionBodyAll(&a[i], b[i], b, bodies, EPS2, KAPPA);
        #pragma omp parallel for
        for(int i = 0; i < bodies; i++)
            updateBody(&b[i], &v[i], a[i], TIMESTEP);            
    }
    dt = omp_get_wtime() - dt;
    
    nBody_state(b, bodies);

    #pragma omp parallel 
    #pragma omp master    
    printf("Threads = %d, CPUs = %d, ", omp_get_num_threads(), omp_get_num_procs());
    printf("Time = %lf, Time per time-step = %lf\n\n", dt, dt / timesteps);

    free(b);
    free(v);
    free(a);

    return 0;    
}
