// n-body problem, MPI solution
//
// compile: 
//      module load mpi
//      mpicc nbody-mpi.c -lm -o nbody-mpi
//
// usage:
//      module load mpi
//      ./nbody-mpi <bodies> <time-steps>
//      
//      module load mpi
//      mpirun -np <processes> ./nbody-mpi <bodies> <time-steps>
//
//      module load mpi
//      srun --reservation=fri --constraint=intel --mpi=pmix --nodes=<nodes> --ntasks=<processes> ./nbody-mpi <bodies> <time-steps>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "/usr/include/openmpi-x86_64/mpi.h"
#include "nbody.h"

struct REAL3    // coordinates
{
    REAL x;
    REAL y;
    REAL z;
};

struct REAL4    // body (coordinates + mass) 
{
    REAL x;     // coordinates
    REAL y;
    REAL z;
    REAL w;     // mass
};

int interactionBodyBody(struct REAL3 *ai, struct REAL4 bi, struct REAL4 bj, REAL eps2, REAL kappa) 
{
    struct REAL3 r;
    
    r.x = bi.x - bj.x;
    r.y = bi.y - bj.y;
    r.z = bi.z - bj.z;

    REAL r2 = r.x * r.x + r.y * r.y + r.z * r.z + eps2; // r^2
    REAL rinv = 1/sqrt(r2);                             // r^{-1}
	REAL a_rinv = kappa * bj.w * rinv * rinv * rinv;    // a r^{-1}

    (*ai).x += a_rinv * r.x;
    (*ai).y += a_rinv * r.y;
    (*ai).z += a_rinv * r.z;

    return 0;
}

int interactionBodyAll(struct REAL3 *ai, struct REAL4 bi, struct REAL4 *b, int bodies, REAL eps2, REAL kappa)
{
    (*ai).x = (*ai).y = (*ai).z = 0.0;

    for (int j = 0; j < bodies; j++)
        interactionBodyBody(ai, bi, b[j], eps2, kappa);

    return 0;
}

int updateBody(struct REAL4 *bi, struct REAL3 *vi, struct REAL3 ai, REAL dt)
{
    (*bi).x += (*vi).x * dt + 0.5 * ai.x * dt * dt;
    (*bi).y += (*vi).y * dt + 0.5 * ai.y * dt * dt;
    (*bi).z += (*vi).z * dt + 0.5 * ai.z * dt * dt;

    (*vi).x += ai.x * dt;
    (*vi).y += ai.y * dt;
    (*vi).z += ai.z * dt;

    return 0;
}

int nBody_init(struct REAL4 *b, struct REAL3 *v, int bodies)
{
    srand(123456);
    for(int i = 0; i < bodies; i++)
    {
        b[i].w = 1.0;
        b[i].x = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        b[i].y = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        b[i].z = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        v[i].x = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        v[i].y = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
        v[i].z = -1.0 + 2.0 * rand()/(REAL)RAND_MAX;
    }
}

int nBody_state(struct REAL4 *b, int bodies)
{
    struct REAL3 pos;
    
    pos.x = 0.0;
    pos.y = 0.0;
    pos.z = 0.0;

    for (int i = 0; i < bodies; i++)
    {
        pos.x += b[i].x;
        pos.y += b[i].y;
        pos.z += b[i].z;
    }
    pos.x /= bodies;
    pos.y /= bodies;
    pos.z /= bodies;

    printf("Position of %d bodies: x = %lf, y = %lf, z = %lf\n", bodies, pos.x, pos.y, pos.z);
}

int main(int argc, char * argv[])
{
    // Init
    int myid, procs;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &myid);
    MPI_Comm_size(MPI_COMM_WORLD, &procs);

    if (argc < 3)
    {
        printf("%s <bodies> <time-steps>\n", argv[0]);
        MPI_Finalize();
        return 1;
    }

    int bodies = atoi(argv[1]);
    int timesteps = atoi(argv[2]);
    int mybodies = bodies / procs;
    int mybodies_start = myid * mybodies;

    if (bodies % procs)
    {
        printf("<bodies> should be multiple of <procs>\n", argv[0]);
        MPI_Finalize();
        return 1;
    }

    // allocate structures
    struct REAL4 *b = (struct REAL4 *)malloc(bodies * sizeof(struct REAL4));
    struct REAL3 *v = (struct REAL3 *)malloc(bodies * sizeof(struct REAL3));

    struct REAL3 *myv = (struct REAL3 *)malloc(mybodies * sizeof(struct REAL3));
    struct REAL3 *mya = (struct REAL3 *)malloc(mybodies * sizeof(struct REAL3));

    // initialization
    if (myid == 0)
    {
        nBody_init(b, v, bodies);
        nBody_state(b, bodies);
    }

    // simulation
    // send particle data to processes
    double dt = MPI_Wtime();
    MPI_Bcast(b, 4 * bodies, MPI_MYREAL, 0, MPI_COMM_WORLD);
    MPI_Scatter(v, 3 * mybodies, MPI_MYREAL, myv, 3 * mybodies, MPI_MYREAL, 0, MPI_COMM_WORLD);

    double dti = MPI_Wtime();    
    for (int t = 0; t < timesteps; t++)
    {   
        for(int i = 0; i < mybodies; i++)
            interactionBodyAll(&mya[i], b[mybodies_start+i], b, bodies, EPS2, KAPPA);
        for(int i = 0; i < mybodies; i++)
            updateBody(&b[mybodies_start+i], &myv[i], mya[i], TIMESTEP);            
        MPI_Allgather(&b[mybodies_start], 4 * mybodies, MPI_MYREAL, b, 4 * mybodies, MPI_MYREAL, MPI_COMM_WORLD);
    }    
    dti = MPI_Wtime() - dti;
    
    MPI_Gather(myv, 3 * mybodies, MPI_MYREAL, v, 3 * mybodies, MPI_MYREAL, 0, MPI_COMM_WORLD);
    dt = MPI_Wtime() - dt;
    
    if (myid == 0)
    {
        nBody_state(b, bodies);

        printf("Procs = %d, ", procs);
        printf("Time = %lf, Time per time-step = %lf\n\n", dt, dti / timesteps);
    }

    free(b);
    free(v);
    free(myv);    
    free(mya);    

    MPI_Finalize();

    return 0;    
}
